@@ -1,7 +1,7 @@
 ////////////////////////////////////////////////////////////////////////////
 //                           **** WAVPACK ****                            //
 //                  Hybrid Lossless Wavefile Compressor                   //
-//              Copyright (c) 1998 - 2007 Conifer Software.               //
+//              Copyright (c) 1998 - 2009 Conifer Software.               //
 //                          All Rights Reserved.                          //
 //      Distributed under the BSD Software License (see license.txt)      //
 ////////////////////////////////////////////////////////////////////////////
@@ -14,6 +14,11 @@
 #include <windows.h>
 #include <io.h>
 #else
+#if defined(__OS2__)
+#define INCL_DOSPROCESS
+#include <os2.h>
+#include <io.h>
+#endif
 #include <sys/stat.h>
 #include <sys/param.h>
 #include <locale.h>
@@ -61,7 +66,7 @@ static char *strdup (const char *s)
 
 static const char *sign_on = "\n"
 " WVUNPACK  Hybrid Lossless Audio Decompressor  %s Version %s\n"
-" Copyright (c) 1998 - 2007 Conifer Software.  All Rights Reserved.\n\n";
+" Copyright (c) 1998 - 2009 Conifer Software.  All Rights Reserved.\n\n";
 
 static const char *usage =
 #if defined (WIN32)
@@ -72,14 +77,17 @@ static const char *usage =
 "             (infile may contain wildcards: ?,*)\n\n"
 " Options: -b  = blindly decode all stream blocks & ignore length info\n"
 "          -c  = extract cuesheet only to stdout (no audio decode)\n"
+"               (note: equivalent to -x \"cuesheet\")\n"
 "          -cc = extract cuesheet file (.cue) in addition to audio file\n"
+"               (note: equivalent to -xx \"cuesheet=%a.cue\")\n"
 "          -d  = delete source file if successful (use with caution!)\n"
 "          --help = this help display\n"
 "          -i  = ignore .wvc file (forces hybrid lossy decompression)\n"
-#if defined (WIN32)
+#if defined (WIN32) || defined (__OS2__)
 "          -l  = run at low priority (for smoother multitasking)\n"
 #endif
 "          -m  = calculate and display MD5 signature; verify if lossless\n"
+"          --no-utf8-convert = leave tag items in UTF-8 on extract or display\n"
 "          -q  = quiet (keep console output to a minimum)\n"
 #if !defined (WIN32)
 "          -o FILENAME | PATH = specify output filename or path\n"
@@ -90,10 +98,16 @@ static const char *usage =
 "          --skip=[sample|hh:mm:ss.ss] = start decoding at specified sample/time\n"
 "          -t  = copy input file's time stamp to output file(s)\n"
 "          --until=[+|-][sample|hh:mm:ss.ss] = stop decoding at specified sample/time\n"
-"            (specifying a '+' causes sample/time to be relative to '--skip' point;\n"
-"             specifying a '-' causes sample/time to be relative to end of file)\n"
+"              (specifying a '+' causes sample/time to be relative to '--skip' point;\n"
+"               specifying a '-' causes sample/time to be relative to end of file)\n"
 "          -v  = verify source data only (no output file created)\n"
 "          -w  = regenerate .wav header (ignore RIFF data in file)\n"
+"          -x \"Field\" = extract specified tag field only to stdout (no audio decode)\n"
+"          -xx \"Field[=file]\" = extract specified tag field to file, optional\n"
+"              filename specification can inlude following replacement codes:\n"
+"                %a = audio output filename\n"
+"                %t = tag field name (note: comes from data for binary tags)\n"
+"                %e = extension from binary tag source file, or 'txt' for text tag\n"
 "          -y  = yes to overwrite warning (use with caution!)\n\n"
 " Web:     Visit www.wavpack.com for latest version and info\n";
 
@@ -102,7 +116,7 @@ static const char *usage =
 
 int debug_logging_mode;
 
-static char overwrite_all, delete_source, raw_decode, extract_cuesheet,
+static char overwrite_all, delete_source, raw_decode, no_utf8_convert,
     summary, ignore_wvc, quiet_mode, calc_md5, copy_time, blind_decode, wav_decode;
 
 static int num_files, file_index, outbuf_k;
@@ -112,8 +126,13 @@ static struct sample_time_index {
     double value;
 } skip, until;
 
+static char *tag_extract_stdout;    // extract single tag to stdout
+static char **tag_extractions;      // extract multiple tags to named files
+static int num_tag_extractions;
+
 /////////////////////////// local function declarations ///////////////////////
 
+static void add_tag_extraction_to_list (char *spec);
 static void parse_sample_time_index (struct sample_time_index *dst, char *src);
 static int unpack_file (char *infilename, char *outfilename);
 static void display_progress (double file_progress);
@@ -128,7 +147,10 @@ static void display_progress (double file_progress);
 
 int main (argc, argv) int argc; char **argv;
 {
-    int verify_only = 0, error_count = 0, add_extension = 0, output_spec = 0, ask_help = 0;
+#ifdef __EMX__ /* OS/2 */
+    _wildcard (&argc, &argv);
+#endif
+    int verify_only = 0, error_count = 0, add_extension = 0, output_spec = 0, ask_help = 0, c_count = 0, x_count = 0;
     char outpath, **matches = NULL, *outfilename = NULL;
     int result;
 
@@ -171,6 +193,8 @@ int main (argc, argv) int argc; char **argv;
 
             if (!strcmp (long_option, "help"))                          // --help
                 ask_help = 1;
+            else if (!strcmp (long_option, "no-utf8-convert"))          // --no-utf8-convert
+                no_utf8_convert = 1;
             else if (!strncmp (long_option, "skip", 4)) {               // --skip
                 parse_sample_time_index (&skip, long_param);
 
@@ -204,7 +228,11 @@ int main (argc, argv) int argc; char **argv;
                         break;
 
                     case 'C': case 'c':
-                        ++extract_cuesheet;
+                        if (++c_count == 2) {
+                            add_tag_extraction_to_list ("cuesheet=%a.cue");
+                            c_count = 0;
+                        }
+
                         break;
 
                     case 'D': case 'd':
@@ -215,6 +243,14 @@ int main (argc, argv) int argc; char **argv;
                     case 'L': case 'l':
                         SetPriorityClass (GetCurrentProcess(), IDLE_PRIORITY_CLASS);
                         break;
+#elif defined (__OS2__)
+                    case 'L': case 'l':
+                        DosSetPriority (0, PRTYC_IDLETIME, 0, 0);
+                        break;
+#endif
+#if defined (WIN32)
+                    case 'O': case 'o':  // ignore -o in Windows to be Linux compatible
+                        break;
 #else
                     case 'O': case 'o':
                         output_spec = 1;
@@ -257,6 +293,15 @@ int main (argc, argv) int argc; char **argv;
                         quiet_mode = 1;
                         break;
 
+                    case 'X': case 'x':
+                        if (++x_count == 3) {
+                            error_line ("illegal option: %s !", *argv);
+                            ++error_count;
+                            x_count = 0;
+                        }
+
+                        break;
+
                     case 'I': case 'i':
                         ignore_wvc = 1;
                         break;
@@ -266,8 +311,22 @@ int main (argc, argv) int argc; char **argv;
                         ++error_count;
                 }
         else {
+            if (x_count) {
+                if (x_count == 1) {
+                    if (tag_extract_stdout) {
+                        error_line ("can't extract more than 1 tag item to stdout at a time!");
+                        ++error_count;
+                    }
+                    else
+                        tag_extract_stdout = *argv;
+                }
+                else if (x_count == 2)
+                    add_tag_extraction_to_list (*argv);
+
+                x_count = 0;
+            }
 #if defined (WIN32)
-            if (!num_files) {
+            else if (!num_files) {
                 matches = realloc (matches, (num_files + 1) * sizeof (*matches));
                 matches [num_files] = malloc (strlen (*argv) + 10);
                 strcpy (matches [num_files], *argv);
@@ -287,7 +346,7 @@ int main (argc, argv) int argc; char **argv;
                 ++error_count;
             }
 #else
-            if (output_spec) {
+            else if (output_spec) {
                 outfilename = malloc (strlen (*argv) + PATH_MAX);
                 strcpy (outfilename, *argv);
                 output_spec = 0;
@@ -324,8 +383,22 @@ int main (argc, argv) int argc; char **argv;
         ++error_count;
     }
 
-    if (extract_cuesheet > 1 && outfilename && *outfilename == '-') {
-        error_line ("can't extract cuesheet when unpacking to stdout!");
+    if (c_count == 1) {
+        if (tag_extract_stdout) {
+            error_line ("can't extract more than 1 tag item to stdout at a time!");
+            error_count++;
+        }
+        else
+            tag_extract_stdout = "cuesheet";
+    }
+
+    if (tag_extract_stdout && (num_tag_extractions || outfilename || verify_only || delete_source || wav_decode || raw_decode)) {
+        error_line ("can't extract a tag to stdout and do anything else!");
+        ++error_count;
+    }
+
+    if ((tag_extract_stdout || num_tag_extractions) && outfilename && *outfilename == '-') {
+        error_line ("can't extract tags when unpacking audio to stdout!");
         ++error_count;
     }
 
@@ -601,10 +674,11 @@ static void parse_sample_time_index (struct sample_time_index *dst, char *src)
 // the resulting audio data and verifying the sum if a sum was stored in the
 // source and lossless compression is used.
 
+static int do_tag_extractions (WavpackContext *wpc, char *outfilename);
 static uchar *format_samples (int bps, uchar *dst, int32_t *src, uint32_t samcnt);
 static void dump_summary (WavpackContext *wpc, char *name, FILE *dst);
 static int write_riff_header (FILE *outfile, WavpackContext *wpc, uint32_t total_samples);
-static int dump_cuesheet (WavpackContext *wpc, FILE *dst);
+static int dump_tag_item_to_file (WavpackContext *wpc, const char *tag_item, FILE *dst, char *fn);
 
 static int unpack_file (char *infilename, char *outfilename)
 {
@@ -639,7 +713,7 @@ static int unpack_file (char *infilename, char *outfilename)
     if (!ignore_wvc)
         open_flags |= OPEN_WVC;
 
-    if (summary > 1 || extract_cuesheet)
+    if (summary > 1 || num_tag_extractions || tag_extract_stdout)
         open_flags |= OPEN_TAGS;
 
     wpc = WavpackOpenFileInput (infilename, error, open_flags, 0);
@@ -659,7 +733,7 @@ static int unpack_file (char *infilename, char *outfilename)
 
     if (skip.value_is_valid) {
         if (skip.value_is_time)
-            skip_sample_index = (uint32_t) skip.value * WavpackGetSampleRate (wpc);
+            skip_sample_index = (uint32_t) (skip.value * WavpackGetSampleRate (wpc));
         else
             skip_sample_index = (uint32_t) skip.value;
 
@@ -715,9 +789,9 @@ static int unpack_file (char *infilename, char *outfilename)
         return NO_ERROR;
     }
 
-    if (extract_cuesheet == 1) {
-        if (!dump_cuesheet (wpc, stdout)) {
-            error_line ("cuesheet not found!");
+    if (tag_extract_stdout) {
+        if (!dump_tag_item_to_file (wpc, tag_extract_stdout, stdout, NULL)) {
+            error_line ("tag \"%s\" not found!", tag_extract_stdout);
             WavpackCloseFile (wpc);
             return SOFT_ERROR;
         }
@@ -725,53 +799,8 @@ static int unpack_file (char *infilename, char *outfilename)
         WavpackCloseFile (wpc);
         return NO_ERROR;
     }
-    else if (extract_cuesheet > 1 && outfilename && *outfilename != '-' && dump_cuesheet (wpc, NULL)) {
-        char *cuefilename = malloc (strlen (outfilename) + 10);
-
-        strcpy (cuefilename, outfilename);
-
-        if (filespec_ext (cuefilename))
-            strcpy (filespec_ext (cuefilename), ".cue");
-        else
-            strcat (cuefilename, ".cue");
-
-        if (!overwrite_all && (outfile = fopen (cuefilename, "r")) != NULL) {
-            DoCloseHandle (outfile);
-            fprintf (stderr, "overwrite %s (yes/no/all)? ", FN_FIT (cuefilename));
-#if defined(WIN32)
-            SetConsoleTitle ("overwrite?");
-#endif
-            switch (yna ()) {
-
-                case 'n':
-                    result = SOFT_ERROR;
-                    break;
-
-                case 'a':
-                    overwrite_all = 1;
-            }
-        }
-
-        // open output file for writing
-
-        if (result == NO_ERROR) {
-            if ((outfile = fopen (cuefilename, "wt")) == NULL) {
-                error_line ("can't create file %s!", FN_FIT (cuefilename));
-                result = SOFT_ERROR;
-            }
-            else {
-                dump_cuesheet (wpc, outfile);
-
-                if (!DoCloseHandle (outfile)) {
-                    error_line ("can't close file %s!", FN_FIT (cuefilename));
-                    result = SOFT_ERROR;
-                }
-                else if (!quiet_mode)
-                    error_line ("extracted cuesheet file %s", FN_FIT (cuefilename));
-            }
-        }
-
-        free (cuefilename);
+    else if (num_tag_extractions && outfilename && *outfilename != '-' && filespec_name (outfilename)) {
+        result = do_tag_extractions (wpc, outfilename);
 
         if (result != NO_ERROR) {
             WavpackCloseFile (wpc);
@@ -822,6 +851,9 @@ static int unpack_file (char *infilename, char *outfilename)
 #if defined(WIN32)
             _setmode (fileno (stdout), O_BINARY);
 #endif
+#if defined(__OS2__)
+            setmode (fileno (stdout), O_BINARY);
+#endif
 
             if (!quiet_mode)
                 fprintf (stderr, "unpacking %s%s to stdout,", *infilename == '-' ?
@@ -1111,6 +1143,129 @@ static int unpack_file (char *infilename, char *outfilename)
     return result;
 }
 
+static void add_tag_extraction_to_list (char *spec)
+{
+    tag_extractions = realloc (tag_extractions, (num_tag_extractions + 1) * sizeof (*tag_extractions));
+    tag_extractions [num_tag_extractions] = malloc (strlen (spec) + 10);
+    strcpy (tag_extractions [num_tag_extractions], spec);
+    num_tag_extractions++;
+}
+
+static int do_tag_extractions (WavpackContext *wpc, char *outfilename)
+{
+    int result = NO_ERROR, i;
+    FILE *outfile;
+
+    for (i = 0; result == NO_ERROR && i < num_tag_extractions; ++i) {
+        char *output_spec = strchr (tag_extractions [i], '=');
+        char tag_filename [256];
+
+        if (output_spec && output_spec > tag_extractions [i] && strlen (output_spec) > 1)
+            *output_spec++ = 0;
+
+        if (dump_tag_item_to_file (wpc, tag_extractions [i], NULL, tag_filename)) {
+            int max_length = (int) strlen (outfilename) + (int) strlen (tag_filename) + 10;
+            char *full_filename;
+
+            if (output_spec)
+                max_length += (int) strlen (output_spec) + 256;
+
+            full_filename = malloc (max_length * 2 + 1);
+            strcpy (full_filename, outfilename);
+
+            if (output_spec) {
+                char *dst = filespec_name (full_filename);
+
+                while (*output_spec && dst - full_filename < max_length) {
+                    if (*output_spec == '%') {
+                        switch (*++output_spec) {
+                            case 'a':                           // audio filename
+                                strcpy (dst, filespec_name (outfilename));
+
+                                if (filespec_ext (dst))         // get rid of any extension
+                                    dst = filespec_ext (dst);
+                                else
+                                    dst += strlen (dst);
+
+                                output_spec++;
+                                break;
+
+                            case 't':                           // tag field name
+                                strcpy (dst, tag_filename);
+
+                                if (filespec_ext (dst))         // get rid of any extension
+                                    dst = filespec_ext (dst);
+                                else
+                                    dst += strlen (dst);
+
+                                output_spec++;
+                                break;
+
+                            case 'e':                           // default extension
+                                if (filespec_ext (tag_filename)) {
+                                    strcpy (dst, filespec_ext (tag_filename) + 1);
+                                    dst += strlen (dst);
+                                }
+
+                                output_spec++;
+                                break;
+
+                            default:
+                                *dst++ = '%';
+                        }
+                    }
+                    else
+                        *dst++ = *output_spec++;
+                }
+
+                *dst = 0;
+            }
+            else
+                strcpy (filespec_name (full_filename), tag_filename);
+
+            if (!overwrite_all && (outfile = fopen (full_filename, "r")) != NULL) {
+                DoCloseHandle (outfile);
+                fprintf (stderr, "overwrite %s (yes/no/all)? ", FN_FIT (full_filename));
+#if defined(WIN32)
+                SetConsoleTitle ("overwrite?");
+#endif
+                switch (yna ()) {
+
+                    case 'n':
+                        *full_filename = 0;
+                        break;
+
+                    case 'a':
+                        overwrite_all = 1;
+                }
+            }
+
+            // open output file for writing
+
+            if (*full_filename) {
+                if ((outfile = fopen (full_filename, "w")) == NULL) {
+                    error_line ("can't create file %s!", FN_FIT (full_filename));
+                    result = SOFT_ERROR;
+                }
+                else {
+                    dump_tag_item_to_file (wpc, tag_extractions [i], outfile, NULL);
+
+                    if (!DoCloseHandle (outfile)) {
+                        error_line ("can't close file %s!", FN_FIT (full_filename));
+                        result = SOFT_ERROR;
+                    }
+                    else if (!quiet_mode)
+                        error_line ("extracted tag \"%s\" to file %s", tag_extractions [i], FN_FIT (full_filename));
+                }
+            }
+
+            free (full_filename);
+        }
+    }
+
+    return result;
+}
+
 // Reformat samples from longs in processor's native endian mode to
 // little-endian data with (possibly) less than 4 bytes / sample.
 
@@ -1261,7 +1416,7 @@ static void dump_summary (WavpackContext *wpc, char *name, FILE *dst)
         WavpackGetSampleRate (wpc));
 
     if (!channel_mask)
-        strcpy (modes, "undefined speakers");
+        strcpy (modes, "unassigned speakers");
     else if (num_channels == 1 && channel_mask == 0x4)
         strcpy (modes, "mono");
     else if (num_channels == 2 && channel_mask == 0x3)
@@ -1313,12 +1468,23 @@ static void dump_summary (WavpackContext *wpc, char *name, FILE *dst)
     else if (WavpackGetMode (wpc) & MODE_HIGH)
         strcat (modes, ", high");
 
-    if (WavpackGetMode (wpc) & MODE_EXTRA)
+    if (WavpackGetMode (wpc) & MODE_EXTRA) {
         strcat (modes, ", extra");
 
+        if (WavpackGetMode (wpc) & MODE_XMODE) {
+            char xmode[3] = "-0";
+
+            xmode [1] = ((WavpackGetMode (wpc) & MODE_XMODE) >> 12) + '0';
+            strcat (modes, xmode);
+        }
+    }
+
     if (WavpackGetMode (wpc) & MODE_SFX)
         strcat (modes, ", sfx");
 
+    if (WavpackGetMode (wpc) & MODE_DNS)
+        strcat (modes, ", dns");
+
     fprintf (dst, "modalities:        %s\n", modes);
 
     if (WavpackGetRatio (wpc) != 0.0) {
@@ -1372,9 +1538,11 @@ static void dump_summary (WavpackContext *wpc, char *name, FILE *dst)
 
     if (WavpackGetMode (wpc) & MODE_VALID_TAG) {
         int ape_tag = WavpackGetMode (wpc) & MODE_APETAG;
+        int num_binary_items = WavpackGetNumBinaryTagItems (wpc);
         int num_items = WavpackGetNumTagItems (wpc), i;
+        char *spaces = "                  ";
 
-        fprintf (dst, "%s tag items:   %d\n\n", ape_tag ? "APEv2" : "ID3v1", num_items);
+        fprintf (dst, "\n%s tag items:   %d\n", ape_tag ? "APEv2" : "ID3v1", num_items + num_binary_items);
 
         for (i = 0; i < num_items; ++i) {
             int item_len, value_len, j;
@@ -1387,40 +1555,159 @@ static void dump_summary (WavpackContext *wpc, char *name, FILE *dst)
             value = malloc (value_len * 2 + 1);
             WavpackGetTagItem (wpc, item, value, value_len + 1);
 
+            fprintf (dst, "%s:%s", item, strlen (item) < strlen (spaces) ? spaces + strlen (item) : " ");
+
             if (ape_tag) {
                 for (j = 0; j < value_len; ++j)
                     if (!value [j])
                         value [j] = '\\';
 
-                fprintf (dst, "%s =%c", item, strchr (value, '\n') ? '\n' : ' ');
-                dump_UTF8_string (value, dst);
-                fprintf (dst, "\n");
+                if (strchr (value, '\n'))
+                    fprintf (dst, "%d-byte multi-line text string\n", value_len);
+                else {
+                    dump_UTF8_string (value, dst);
+                    fprintf (dst, "\n");
+                }
             }
             else
-                fprintf (dst, "%s = %s\n", item, value);
+                fprintf (dst, "%s\n", value);
 
             free (value);
             free (item);
         }
+
+        for (i = 0; i < num_binary_items; ++i) {
+            int item_len, value_len;
+            char *item, fname [256];
+
+            item_len = WavpackGetBinaryTagItemIndexed (wpc, i, NULL, 0);
+            item = malloc (item_len + 1);
+            WavpackGetBinaryTagItemIndexed (wpc, i, item, item_len + 1);
+            value_len = dump_tag_item_to_file (wpc, item, NULL, fname);
+            fprintf (dst, "%s:%s", item, strlen (item) < strlen (spaces) ? spaces + strlen (item) : " ");
+
+            if (filespec_ext (fname))
+                fprintf (dst, "%d-byte binary item (%s)\n", value_len, filespec_ext (fname)+1);
+            else
+                fprintf (dst, "%d-byte binary item\n", value_len);
+
+#if 0   // debug binary tag reading
+            {
+                char md5_string [] = "00000000000000000000000000000000";
+                uchar md5_result [16];
+                MD5_CTX md5_context;
+                char *value;
+                int i, j;
+
+                MD5Init (&md5_context);
+                value_len = WavpackGetBinaryTagItem (wpc, item, NULL, 0);
+                value = malloc (value_len);
+                value_len = WavpackGetBinaryTagItem (wpc, item, value, value_len);
+
+                for (i = 0; i < value_len; ++i)
+                    if (!value [i]) {
+                        MD5Update (&md5_context, (unsigned char *) value + i + 1, value_len - i - 1);
+                        MD5Final (md5_result, &md5_context);
+                        for (j = 0; j < 16; ++j)
+                            sprintf (md5_string + (j * 2), "%02x", md5_result [j]);
+                        fprintf (dst, "    %d byte string >>%s<<\n", i, value);
+                        fprintf (dst, "    %d bytes binary data >>%s<<\n", value_len - i - 1, md5_string);
+                        break;
+                    }
+
+                if (i == value_len)
+                    fprintf (dst, "    no NULL found in binary value (or value not readable)\n");
+
+                free (value);
+            }
+#endif
+            free (item);
+        }
     }
 }
 
-static int dump_cuesheet (WavpackContext *wpc, FILE *dst)
-{
-    if ((WavpackGetMode (wpc) & MODE_VALID_TAG) &&
-        (WavpackGetMode (wpc) & MODE_APETAG)) {
+// Dump the specified tag field to the specified stream. Both text and binary tags may be written,
+// and in Windows the appropriate file mode will be set. If the tag is not found then 0 is returned,
+// otherwise the length of the data is returned, and this is true even when the file pointer is NULL
+// so this can be used to determine if the tag exists before further processing.
+//
+// The "fname" parameter can optionally be set to a character array that will accept the suggested
+// filename. This is formed by the tag item name with the extension ".txt" for text fields; for
+// binary fields this is supplied by convention as a NULL terminated string at the beginning of the
+// data, so this is returned. The string should have 256 characters available.
 
-            int value_len = WavpackGetTagItem (wpc, "cuesheet", NULL, 0);
+static int dump_tag_item_to_file (WavpackContext *wpc, const char *tag_item, FILE *dst, char *fname)
+{
+    if (WavpackGetMode (wpc) & MODE_VALID_TAG) {
+        if (WavpackGetTagItem (wpc, tag_item, NULL, 0)) {
+            int value_len = WavpackGetTagItem (wpc, tag_item, NULL, 0);
             char *value;
 
+            if (fname) {
+                strcpy (fname, tag_item);
+                strcat (fname, ".txt");
+            }
+
             if (!value_len || !dst)
                 return value_len;
 
+#if defined(WIN32)
+            _setmode (fileno (dst), O_TEXT);
+#endif
+#if defined(__OS2__)
+            setmode (fileno (dst), O_TEXT);
+#endif
             value = malloc (value_len * 2 + 1);
-            WavpackGetTagItem (wpc, "cuesheet", value, value_len + 1);
+            WavpackGetTagItem (wpc, tag_item, value, value_len + 1);
             dump_UTF8_string (value, dst);
             free (value);
             return value_len;
+        }
+        else if (WavpackGetBinaryTagItem (wpc, tag_item, NULL, 0)) {
+            int value_len = WavpackGetBinaryTagItem (wpc, tag_item, NULL, 0), res = 0, i;
+            uint32_t bcount = 0;
+            char *value;
+
+            value = malloc (value_len);
+            WavpackGetBinaryTagItem (wpc, tag_item, value, value_len);
+
+            for (i = 0; i < value_len; ++i)
+                if (!value [i]) {
+
+                    if (dst) {
+#if defined(WIN32)
+                        _setmode (fileno (dst), O_BINARY);
+#endif
+#if defined(__OS2__)
+                        setmode (fileno (dst), O_BINARY);
+#endif
+                        res = DoWriteFile (dst, (unsigned char *) value + i + 1, value_len - i - 1, &bcount);
+                    }
+
+                    if (fname) {
+                        if (i < 256)
+                            strcpy (fname, value);
+                        else {
+                            strcpy (fname, tag_item);
+                            strcat (fname, ".bin");
+                        }
+                    }
+
+                    break;
+                }
+
+            free (value);
+
+            if (i == value_len)
+                return 0;
+
+            if (dst && (!res || bcount != value_len - i - 1))
+                return 0;
+
+            return value_len - i - 1;
+        }
+        else
+            return 0;
     }
     else
         return 0;
@@ -1460,7 +1747,10 @@ static void dump_UTF8_string (char *string, FILE *dst)
         }
 
         *p = 0;
-        UTF8ToAnsi (temp, len * 2);
+
+        if (!no_utf8_convert)
+            UTF8ToAnsi (temp, len * 2);
+
         fputs (temp, dst);
         free (temp);
     }
@@ -1547,12 +1837,19 @@ static void UTF8ToAnsi (char *string, int len)
     size_t outsize = len - 1;
     int err = 0;
     char *old_locale;
+    iconv_t converter;
 
     memset(temp, 0, len);
     old_locale = setlocale (LC_CTYPE, "");
-    iconv_t converter = iconv_open ("", "UTF-8");
-    err = iconv (converter, &inp, &insize, &outp, &outsize);
-    iconv_close (converter);
+    converter = iconv_open ("", "UTF-8");
+
+    if (converter != (iconv_t) -1) {
+        err = iconv (converter, &inp, &insize, &outp, &outsize);
+        iconv_close (converter);
+    }
+    else
+        err = -1;
+
     setlocale (LC_CTYPE, old_locale);
 
     if (err == -1) {
