@@ -64,7 +64,13 @@ int unpack_init (WavpackContext *wpc)
     CLEAR (wps->dc);
     CLEAR (wps->w);
 
-    if (wps->wphdr.flags & UNKNOWN_FLAGS) {
+    if (!(wps->wphdr.flags & MONO_FLAG) && wpc->config.num_channels && wps->wphdr.block_samples &&
+        (wpc->reduced_channels == 1 || wpc->config.num_channels == 1)) {
+            wps->mute_error = TRUE;
+            return FALSE;
+    }
+
+    if ((wps->wphdr.flags & UNKNOWN_FLAGS) || (wps->wphdr.flags & MONO_DATA) == MONO_DATA) {
         wps->mute_error = TRUE;
         return FALSE;
     }
@@ -115,6 +121,9 @@ int unpack_init (WavpackContext *wpc)
 
 int init_wv_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
+    if (!wpmd->byte_length)
+        return FALSE;
+
     bs_open_read (&wps->wvbits, wpmd->data, (uchar *) wpmd->data + wpmd->byte_length);
     return TRUE;
 }
@@ -124,6 +133,9 @@ int init_wv_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 
 int init_wvc_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
+    if (!wpmd->byte_length)
+        return FALSE;
+
     bs_open_read (&wps->wvcbits, wpmd->data, (uchar *) wpmd->data + wpmd->byte_length);
     return TRUE;
 }
@@ -139,6 +151,9 @@ int init_wvx_bitstream (WavpackStream *wps, WavpackMetadata *wpmd)
 {
     uchar *cp = wpmd->data;
 
+    if (wpmd->byte_length <= 4)
+        return FALSE;
+
     wps->crc_wvx = *cp++;
     wps->crc_wvx |= (int32_t) *cp++ << 8;
     wps->crc_wvx |= (int32_t) *cp++ << 16;
@@ -229,6 +244,9 @@ int read_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
     }
 
     if (wps->wphdr.version == 0x402 && (wps->wphdr.flags & HYBRID_FLAG)) {
+        if (byteptr + (wps->wphdr.flags & MONO_DATA ? 2 : 4) > endptr)
+            return FALSE;
+
         wps->dc.error [0] = exp2s ((short)(byteptr [0] + (byteptr [1] << 8)));
         byteptr += 2;
 
@@ -240,6 +258,9 @@ int read_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
 
     while (dpp-- > wps->decorr_passes && byteptr < endptr)
         if (dpp->term > MAX_TERM) {
+            if (byteptr + (wps->wphdr.flags & MONO_DATA ? 4 : 8) > endptr)
+                return FALSE;
+
             dpp->samples_A [0] = exp2s ((short)(byteptr [0] + (byteptr [1] << 8)));
             dpp->samples_A [1] = exp2s ((short)(byteptr [2] + (byteptr [3] << 8)));
             byteptr += 4;
@@ -251,6 +272,9 @@ int read_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
             }
         }
         else if (dpp->term < 0) {
+            if (byteptr + 4 > endptr)
+                return FALSE;
+
             dpp->samples_A [0] = exp2s ((short)(byteptr [0] + (byteptr [1] << 8)));
             dpp->samples_B [0] = exp2s ((short)(byteptr [2] + (byteptr [3] << 8)));
             byteptr += 4;
@@ -259,6 +283,9 @@ int read_decorr_samples (WavpackStream *wps, WavpackMetadata *wpmd)
             int m = 0, cnt = dpp->term;
 
             while (cnt--) {
+                if (byteptr + (wps->wphdr.flags & MONO_DATA ? 2 : 4) > endptr)
+                    return FALSE;
+
                 dpp->samples_A [m] = exp2s ((short)(byteptr [0] + (byteptr [1] << 8)));
                 byteptr += 2;
 
@@ -343,20 +370,41 @@ int read_int32_info (WavpackStream *wps, WavpackMetadata *wpmd)
 int read_channel_info (WavpackContext *wpc, WavpackMetadata *wpmd)
 {
     int bytecnt = wpmd->byte_length, shift = 0;
-    char *byteptr = wpmd->data;
+    uchar *byteptr = wpmd->data;
     uint32_t mask = 0;
 
-    if (!bytecnt || bytecnt > 5)
+    if (!bytecnt || bytecnt > 6)
         return FALSE;
 
-    wpc->config.num_channels = *byteptr++;
+    if (!wpc->config.num_channels) {
 
-    while (--bytecnt) {
-        mask |= (uint32_t) *byteptr++ << shift;
-        shift += 8;
+        if (bytecnt == 6) {
+            wpc->config.num_channels = (byteptr [0] | ((byteptr [2] & 0xf) << 8)) + 1;
+            wpc->max_streams = (byteptr [1] | ((byteptr [2] & 0xf0) << 4)) + 1;
+
+            if (wpc->config.num_channels < wpc->max_streams)
+                return FALSE;
+    
+            byteptr += 3;
+            mask = *byteptr++;
+            mask |= (uint32_t) *byteptr++ << 8;
+            mask |= (uint32_t) *byteptr << 16;
+        }
+        else {
+            wpc->config.num_channels = *byteptr++;
+
+            while (--bytecnt) {
+                mask |= (uint32_t) *byteptr++ << shift;
+                shift += 8;
+            }
+        }
+
+        if (wpc->config.num_channels > wpc->max_streams * 2)
+            return FALSE;
+
+        wpc->config.channel_mask = mask;
     }
 
-    wpc->config.channel_mask = mask;
     return TRUE;
 }
 
@@ -371,7 +419,10 @@ int read_config_info (WavpackContext *wpc, WavpackMetadata *wpmd)
         wpc->config.flags &= 0xff;
         wpc->config.flags |= (int32_t) *byteptr++ << 8;
         wpc->config.flags |= (int32_t) *byteptr++ << 16;
-        wpc->config.flags |= (int32_t) *byteptr << 24;
+        wpc->config.flags |= (int32_t) *byteptr++ << 24;
+
+        if (bytecnt >= 4 && (wpc->config.flags & CONFIG_EXTRA_MODE))
+            wpc->config.xmode = *byteptr;
     }
 
     return TRUE;
@@ -449,7 +500,11 @@ int32_t unpack_samples (WavpackContext *wpc, int32_t *buffer, uint32_t sample_co
         sample_count = wps->wphdr.block_index + wps->wphdr.block_samples - wps->sample_index;
 
     if (wps->mute_error) {
-        memset (buffer, 0, sample_count * (flags & MONO_FLAG ? 4 : 8));
+        if (wpc->reduced_channels == 1 || wpc->config.num_channels == 1 || (flags & MONO_FLAG))
+            memset (buffer, 0, sample_count * 4);
+        else
+            memset (buffer, 0, sample_count * 8);
+
         wps->sample_index += sample_count;
         return sample_count;
     }
