@@ -88,11 +88,16 @@ typedef struct {
 
 #define APE_Tag_Hdr_Format "8LLLL"
 
+#define APE_TAG_TYPE_TEXT       0x0
+#define APE_TAG_TYPE_BINARY     0x1
+#define APE_TAG_THIS_IS_HEADER  0x20000000
+#define APE_TAG_CONTAINS_HEADER 0x80000000
+
 typedef struct {
     int32_t tag_file_pos;
     ID3_Tag id3_tag;
     APE_Tag_Hdr ape_tag_hdr;
-    char *ape_tag_data;
+    unsigned char *ape_tag_data;
 } M_Tag;
 
 // RIFF / wav header formats (these occur at the beginning of both wav files
@@ -176,7 +181,7 @@ typedef struct {
 
 #define MIN_STREAM_VERS     0x402       // lowest stream version we'll decode
 #define MAX_STREAM_VERS     0x410       // highest stream version we'll decode or encode
-#define CUR_STREAM_VERS     0x406       // stream version we are [normally] writing now
+#define CUR_STREAM_VERS     0x407       // stream version we are [normally] writing now
 
 
 //////////////////////////// WavPack Metadata /////////////////////////////////
@@ -248,6 +253,7 @@ typedef struct {
 #define CONFIG_AUTO_SHAPING     0x4000  // automatic noise shaping
 #define CONFIG_SHAPE_OVERRIDE   0x8000  // shaping mode specified
 #define CONFIG_JOINT_OVERRIDE   0x10000 // joint-stereo mode specified
+#define CONFIG_DYNAMIC_SHAPING  0x20000 // dynamic noise shaping
 #define CONFIG_CREATE_EXE       0x40000 // create executable
 #define CONFIG_CREATE_WVC       0x80000 // create correction file
 #define CONFIG_OPTIMIZE_WVC     0x100000 // maximize bybrid compression
@@ -256,12 +262,18 @@ typedef struct {
 #define CONFIG_EXTRA_MODE       0x2000000 // extra processing mode
 #define CONFIG_SKIP_WVX         0x4000000 // no wvx stream w/ floats & big ints
 #define CONFIG_MD5_CHECKSUM     0x8000000 // compute & store MD5 signature
+#define CONFIG_MERGE_BLOCKS     0x10000000 // merge blocks of equal redundancy (for lossyWAV)
+#define CONFIG_PAIR_UNDEF_CHANS 0x20000000 // encode undefined channels in stereo pairs
 #define CONFIG_OPTIMIZE_MONO    0x80000000 // optimize for mono streams posing as stereo
 
 /*
- * These config flags are no longer used (or were never used) although there
- * may be WavPack files that have some of these bits set in the config
- * metadata structure, so be careful reusing them for something else.
+ * These config flags were never actually used, or are no longer used, or are
+ * used for something else now. They may be used in the future for what they
+ * say, or for something else. WavPack files in the wild *may* have some of
+ * these bit set in their config flags (with these older meanings), but only
+ * if the stream version is 0x410 or less than 0x407. Of course, this is not
+ * very important because once the file has been encoded, the config bits are
+ * just for information purposes (i.e., they do not affect decoding),
  *
 #define CONFIG_ADOBE_MODE       0x100   // "adobe" mode for 32-bit floats
 #define CONFIG_VERY_FAST_FLAG   0x400   // double fast
@@ -305,7 +317,8 @@ typedef struct bs {
 } Bitstream;
 
 #define MAX_WRAPPER_BYTES 16777216
-#define MAX_STREAMS 8
+#define NEW_MAX_STREAMS 4096
+#define OLD_MAX_STREAMS 8
 #define MAX_NTERMS 16
 #define MAX_TERM 8
 
@@ -352,10 +365,12 @@ typedef struct {
     struct {
         int32_t shaping_acc [2], shaping_delta [2], error [2];
         double noise_sum, noise_ave, noise_max;
+        short *shaping_data, *shaping_array;
+        int32_t shaping_samples;
     } dc;
 
-    struct decorr_pass decorr_passes [MAX_NTERMS];
-    WavpackDecorrSpec *decorr_specs;
+    struct decorr_pass decorr_passes [MAX_NTERMS], analysis_pass;
+    const WavpackDecorrSpec *decorr_specs;
 } WavpackStream;
 
 // flags for float_flags:
@@ -406,12 +421,12 @@ typedef struct {
 
     uint32_t filelen, file2len, filepos, file2pos, total_samples, crc_errors, first_flags;
     int wvc_flag, open_flags, norm_offset, reduced_channels, lossy_blocks, close_files;
-    uint32_t block_samples, max_samples, acc_samples, initial_index;
+    uint32_t block_samples, ave_block_samples, block_boundary, max_samples, acc_samples, initial_index, riff_trailer_bytes;
     int riff_header_added, riff_header_created;
     M_Tag m_tag;
 
-    int current_stream, num_streams, stream_version;
-    WavpackStream *streams [MAX_STREAMS];
+    int current_stream, num_streams, max_streams, stream_version;
+    WavpackStream **streams;
     void *stream3;
 
     char error_message [80];
@@ -421,15 +436,26 @@ typedef struct {
 
 #define CLEAR(destin) memset (&destin, 0, sizeof (destin));
 
-// these macros implement the weight application and update operations
-// that are at the heart of the decorrelation loops
+// These macros implement the weight application and update operations
+// that are at the heart of the decorrelation loops. Note that there are
+// sometimes two and even three versions of each macro. Theses should be
+// equivalent and produce identical results, but some may perform better
+// or worse on a given architecture.
 
+#if 1   // PERFCOND - apply decorrelation weight when no 32-bit overflow possible
 #define apply_weight_i(weight, sample) ((weight * sample + 512) >> 10)
+#else
+#define apply_weight_i(weight, sample) ((((weight * sample) >> 8) + 2) >> 2)
+#endif
 
+#if 1   // PERFCOND - apply decorrelation weight when 32-bit overflow is possible
 #define apply_weight_f(weight, sample) (((((sample & 0xffff) * weight) >> 9) + \
     (((sample & ~0xffff) >> 9) * weight) + 1) >> 1)
+#else
+#define apply_weight_f(weight, sample) ((int32_t)floor(((double) weight * sample + 512.0) / 1024.0))
+#endif
 
-#if 1   // PERFCOND
+#if 1   // PERFCOND - universal version that checks input magnitude (or simply uses 64-bit ints)
 #define apply_weight(weight, sample) (sample != (short) sample ? \
     apply_weight_f (weight, sample) : apply_weight_i (weight, sample))
 #else
@@ -650,12 +676,15 @@ int WavpackGetMode (WavpackContext *wpc);
 #define MODE_VALID_TAG  0x10
 #define MODE_HIGH       0x20
 #define MODE_FAST       0x40
-#define MODE_EXTRA      0x80
+#define MODE_EXTRA      0x80    // extra mode used, see MODE_XMODE for possible level
 #define MODE_APETAG     0x100
 #define MODE_SFX        0x200
 #define MODE_VERY_HIGH  0x400
 #define MODE_MD5        0x800
+#define MODE_XMODE      0x7000  // mask for extra level (1-6, 0=unknown)
+#define MODE_DNS        0x8000
 
+char *WavpackGetErrorMessage (WavpackContext *wpc);
 int WavpackGetVersion (WavpackContext *wpc);
 uint32_t WavpackUnpackSamples (WavpackContext *wpc, int32_t *buffer, uint32_t samples);
 uint32_t WavpackGetNumSamples (WavpackContext *wpc);
@@ -670,6 +699,7 @@ int WavpackGetBytesPerSample (WavpackContext *wpc);
 int WavpackGetNumChannels (WavpackContext *wpc);
 int WavpackGetChannelMask (WavpackContext *wpc);
 int WavpackGetReducedChannels (WavpackContext *wpc);
+int WavpackGetFloatNormExp (WavpackContext *wpc);
 int WavpackGetMD5Sum (WavpackContext *wpc, uchar data [16]);
 uint32_t WavpackGetWrapperBytes (WavpackContext *wpc);
 uchar *WavpackGetWrapperData (WavpackContext *wpc);
@@ -680,12 +710,6 @@ uint32_t WavpackGetFileSize (WavpackContext *wpc);
 double WavpackGetRatio (WavpackContext *wpc);
 double WavpackGetAverageBitrate (WavpackContext *wpc, int count_wvc);
 double WavpackGetInstantBitrate (WavpackContext *wpc);
-int WavpackGetNumTagItems (WavpackContext *wpc);
-int WavpackGetTagItem (WavpackContext *wpc, const char *item, char *value, int size);
-int WavpackGetTagItemIndexed (WavpackContext *wpc, int index, char *item, int size);
-int WavpackAppendTagItem (WavpackContext *wpc, const char *item, const char *value, int vsize);
-int WavpackDeleteTagItem (WavpackContext *wpc, const char *item);
-int WavpackWriteTag (WavpackContext *wpc);
 
 WavpackContext *WavpackOpenFileOutput (WavpackBlockOutput blockout, void *wv_id, void *wvc_id);
 int WavpackSetConfiguration (WavpackContext *wpc, WavpackConfig *config, uint32_t total_samples);
@@ -700,6 +724,25 @@ void *WavpackGetWrapperLocation (void *first_block, uint32_t *size);
 void WavpackLittleEndianToNative (void *data, char *format);
 void WavpackNativeToLittleEndian (void *data, char *format);
 
+uint32_t WavpackGetLibraryVersion (void);
+const char *WavpackGetLibraryVersionString (void);
+
+// tags.c
+
+int WavpackGetNumTagItems (WavpackContext *wpc);
+int WavpackGetTagItem (WavpackContext *wpc, const char *item, char *value, int size);
+int WavpackGetTagItemIndexed (WavpackContext *wpc, int index, char *item, int size);
+int WavpackGetNumBinaryTagItems (WavpackContext *wpc);
+int WavpackGetBinaryTagItem (WavpackContext *wpc, const char *item, char *value, int size);
+int WavpackGetBinaryTagItemIndexed (WavpackContext *wpc, int index, char *item, int size);
+int WavpackAppendTagItem (WavpackContext *wpc, const char *item, const char *value, int vsize);
+int WavpackAppendBinaryTagItem (WavpackContext *wpc, const char *item, const char *value, int vsize);
+int WavpackDeleteTagItem (WavpackContext *wpc, const char *item);
+int WavpackWriteTag (WavpackContext *wpc);
+int load_tag (WavpackContext *wpc);
+void free_tag (M_Tag *m_tag);
+int valid_tag (M_Tag *m_tag);
+
 ///////////////////////////// SIMD helper macros /////////////////////////////
 
 #ifdef OPT_MMX
@@ -719,19 +762,28 @@ typedef int __v4hi __attribute__ ((__mode__ (__V4HI__)));
 #define _m_pxor(m1, m2) (__m64) __builtin_ia32_pxor ((__di) m1, (__di) m2)
 #else
 typedef int __m64 __attribute__ ((__vector_size__ (8)));
-#define _m_paddsw(m1, m2) __builtin_ia32_paddsw (m1, m2)
+typedef short __m64_16 __attribute__ ((__vector_size__ (8)));
+#define _m_paddsw(m1, m2) (__m64) __builtin_ia32_paddsw ((__m64_16) m1, (__m64_16) m2)
 #define _m_pand(m1, m2) __builtin_ia32_pand (m1, m2)
 #define _m_pandn(m1, m2) __builtin_ia32_pandn (m1, m2)
-#define _m_pmaddwd(m1, m2) __builtin_ia32_pmaddwd (m1, m2)
+#define _m_pmaddwd(m1, m2) __builtin_ia32_pmaddwd ((__m64_16) m1, (__m64_16) m2)
 #define _m_por(m1, m2) __builtin_ia32_por (m1, m2)
 #define _m_pxor(m1, m2) __builtin_ia32_pxor (m1, m2)
 #endif
 
 #define _m_paddd(m1, m2) __builtin_ia32_paddd (m1, m2)
 #define _m_pcmpeqd(m1, m2) __builtin_ia32_pcmpeqd (m1, m2)
-#define _m_pslldi(m1, m2) __builtin_ia32_pslld (m1, m2)
-#define _m_psradi(m1, m2) __builtin_ia32_psrad (m1, m2)
-#define _m_psrldi(m1, m2) __builtin_ia32_psrld (m1, m2)
+
+#if (__GNUC__ == 4 && __GNUC_MINOR__ >= 4) || __GNUC__ > 4
+#	define _m_pslldi(m1, m2) __builtin_ia32_pslldi ((__m64)m1, m2)
+#	define _m_psradi(m1, m2) __builtin_ia32_psradi ((__m64)m1, m2)
+#	define _m_psrldi(m1, m2) __builtin_ia32_psrldi ((__m64)m1, m2)
+#else
+#	define _m_pslldi(m1, m2) __builtin_ia32_pslld (m1, m2)
+#	define _m_psradi(m1, m2) __builtin_ia32_psrad (m1, m2)
+#	define _m_psrldi(m1, m2) __builtin_ia32_psrld (m1, m2)
+#endif
+
 #define _m_psubd(m1, m2) __builtin_ia32_psubd (m1, m2)
 #define _m_punpckhdq(m1, m2) __builtin_ia32_punpckhdq (m1, m2)
 #define _m_punpckldq(m1, m2) __builtin_ia32_punpckldq (m1, m2)
